(*
 * lTerm_widget.ml
 * ---------------
 * Copyright : (c) 2019, ZAN DoYe <zandoye@gmail.com>
 * Licence   : BSD3
 *
 * This file is a part of Lambda-Term.
 *)

include LTerm_text_impl.Make(LiteralIntf.UTF8)

