(*
 * lTerm_editor.mli
 * ------------
 * Copyright : (c) 2020, ZAN DoYe <zandoye@gmail.com>
 * Licence   : BSD3
 *
 * This file is a part of Lambda-Term.
 *)

(** The editor mode *)
type mode=
  | Default
  | Vi


